﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Globalization;
using System.Reflection;
using System.Text;
using Produmex.Foundation;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.Messages;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf;
using Produmex.Foundation.Wwf.Sbo;
using Produmex.Foundation.Wwf.Sbo.DataObjects;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Providers;

namespace WorkflowScript_ImpermoScanSSCCOnLocationScript
{
    /// <summary>
    /// Impermo
    /// </summary>
    public class WorkflowScript_ImpermoScanSSCCOnLocationScript : WorkflowInstanceScriptBase
    {
        /// <summary>
        /// Definitions
        /// </summary>
        public static class Definitions
        {
            /// <summary>
            /// Location to store SSCC's not on counted location
            /// </summary>
            public const string TEMP_LOCATION_CODE = "";
        }

        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);

        // Input parameters * do not change *
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;

        // Output parameters

        // Sub-flows
        // <none>



        /// <summary>
        /// Initializes a new instance of the <see cref="WorkflowScript_ImpermoScanSSCCOnLocationScript"/> class.
        /// </summary>
        /// <param name="parent">The parent.</param>
        /// <param name="factory">The factory.</param>
        public WorkflowScript_ImpermoScanSSCCOnLocationScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
            : base(parent, factory)
        {
        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {
            // Parameters in scope
            Session session = GetScopeParameter("Session") as Session;
            ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;

            string initialErrorKey = null;
            Message msg = null;

            string query = null;
           
            DataSet dsLocations = null;
            string locationCode = null;
            string enteredLocation = null;


            DataSet dsSSCC = null;
            Collection<int> listOfLUIDOnOtherLocation = null;
            Collection<int> listOfLUID = null; ;


            SboUserInfo userInfo = sboProviderService.GetCurrentUser();

        Step_IdentifyLocation:

            #region IDENTIFY LOCATION

            enteredLocation = null;
            locationCode = null;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyLocationScreen),
            DefaultCultureInfo.Get(), BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_LOCATION",
                "HideDefaultLocation", true,
                "AllowedButtons", new string[] { "MSG_BUTTON_SELECT_LOCATION"}));
            msg = WaitForMessage();

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyLocationScreenReply.ButtonClicked")
            {
                goto Step_SelectLocation;
            }

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyLocationScreenReply.LocationEntered")
            {
                enteredLocation = ExtractParameter<string>(msg.Parameters, "locationCode");

                goto Step_SelectLocation;
            }

            //go back
            return;

            #endregion

        Step_SelectLocation:

            #region SELECT LOCATION

            query = BuildQuery.GetAllLocation(null, enteredLocation);

            dsLocations = sboProviderService.RunView(false, null, null, query);

            if (GetNumberOfRows(dsLocations) > 0)
            {
                //we got a location
                if (GetNumberOfRows(dsLocations) == 1)
                {
                    //only one location, so we can take this one
                    locationCode = dsLocations.Tables[0].Rows[0][LocationDataSet.Columns.CODE.Name] as string;

                    goto Step_StartSCCC;
                }
                else
                {
                    //more than one location, so the user can select the location
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "InitialErrorKey", initialErrorKey,
                        "TitleKey", "MSG_TITLE_SELECT_LOCATION",
                        "LocationDS", dsLocations));
                    msg = WaitForMessage();

                    if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.ISelectLocationScreenReply.LocationSelected")
                    {
                        locationCode = ExtractParameter<string>(msg.Parameters, "locationCode");
                        goto Step_StartSCCC;
                    }

                    goto Step_IdentifyLocation;
                }
            }
            else
            {
                //show message screen
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    DefaultCultureInfo.Get(),
                    BuildParamCollection(
                        "MessageKey", "MSG_ERROR_NO_LOCATION_FOUND",
                        "ShowButton", true));
                msg = WaitForMessage();

                goto Step_IdentifyLocation;
            }

            #endregion

        Step_StartSCCC:
            #region START SSCC

            dsSSCC = GetNewSSCCDataSet();
            listOfLUIDOnOtherLocation = new Collection<int>();
            listOfLUID = new Collection<int>();

            goto Step_ScanSSCC;

            #endregion




        Step_ScanSSCC:

            #region SCAN SSCC

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyMultipleSSCCScreen),
                DefaultCultureInfo.Get(), BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "Scan SSCC op locatie",
                    "SSCCDS", dsSSCC));
            msg = WaitForMessage();

            //clear the error
            initialErrorKey = null;

            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyMultipleSSCCScreenReply.BackRequested")
            {
                goto Step_IdentifyLocation;
            }
            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyMultipleSSCCScreenReply.SSCCScanned")
            {
                string sscc = ExtractParameter<string>(msg.Parameters, "sscc");

                DataRow[] rows = dsSSCC.Tables[0].Select(string.Format("{0} = '{1}'", SSCCDataSet.Columns.SSCC.Name, sscc));
                if (rows.Length > 0)
                {
                    initialErrorKey = "MSG_ERROR_SSCC_ALREADY_SCANNED";
                }
                else
                {
                    //get the data for the sscc
                    query = BuildQuery.GetSSCCInfo(sscc);
                    DataSet dsTemp = sboProviderService.RunView(true, null, null, query);

                    if (GetNumberOfRows(dsTemp) > 0)
                    {
                        foreach (DataRow row in dsTemp.Tables[0].Rows)
                        {
                            //copy into the sscc dataset
                            dsSSCC.Tables[0].ImportRow(row);
                            listOfLUID.Add(Convert.ToInt32(row[SSCCDataSet.Columns.LUID.Name]));

                            if (row[SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name].ToString() != locationCode)
                            {
                                //the sscc needs to be moved to current location
                                listOfLUIDOnOtherLocation.Add(Convert.ToInt32(row[SSCCDataSet.Columns.LUID.Name]));
                            }
                        }
                    }
                    else
                    {
                        initialErrorKey = "MSG_ERROR_NO_PRODUCTS_ON_SSCC";
                    }

                }

                goto Step_ScanSSCC;
            }
            if (msg.Name == "Produmex.Foundation.SlimScreen.Interfaces.MoveOrder.IIdentifyMultipleSSCCScreenReply.DeleteSSCC")
            {
                int luid = ExtractParameter<int>(msg.Parameters, "luid");

                listOfLUIDOnOtherLocation.Remove(luid);
                listOfLUID.Remove(luid);

                DataRow[] rows = dsSSCC.Tables[0].Select(SSCCDataSet.Columns.LUID.Name + "=" + luid);
                foreach (DataRow row in rows)
                {
                    dsSSCC.Tables[0].Rows.Remove(row);
                }
                dsSSCC.AcceptChanges();

                goto Step_ScanSSCC;
            }

            //forward requested, so now adjust stock
            goto Step_AdjustStock;

        Step_AdjustStock:

            #region ADJUST STOCK

            //we need to get all luids not on location

            Collection<int> listOfLuidsNotOnLocation = new Collection<int>();

            query = BuildQuery.GetSSCCInfoNotOnLocation(locationCode, listOfLUID);
            DataSet ds = sboProviderService.RunView(false, null, null, query);

            if (this.GetNumberOfRows(ds) > 0)
            {
                foreach (DataRow row in ds.Tables[0].Rows)
                {
                    listOfLuidsNotOnLocation.Add(Convert.ToInt32(row[SSCCDataSet.Columns.LUID.Name]));
                }
            }

            try
            {
                sboProviderService.InvokeMethodWithDbConnection<object>(true, false, null, null, new DelegateWithPmxDbConnection<object>(ExecuteMakeMoves), locationCode, listOfLUIDOnOtherLocation, listOfLuidsNotOnLocation);
            }
            catch (Exception ex)
            {
                s_log.Error("", ex);
                if (!(ex is ProdumexException))
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "MessageKey", ex.Message,
                            "NoTranslationOfMessageKey", true,
                            "ShowButton", true));
                    WaitForMessage();

                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        DefaultCultureInfo.Get(), BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();
                }

                //ask user if he wants to try again after error
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    DefaultCultureInfo.Get(), BuildParamCollection(
                        "TitleKey", "MSG_TITLE_ERROR_TRY_AGAIN",
                        "MessageKey", "MSG_QUESTION_ERROR_TRY_AGAIN"));
                msg = WaitForMessage();

                if (msg.Name.EndsWith(".Yes"))
                {
                    goto Step_AdjustStock;
                }
            }



            goto Step_IdentifyLocation;
            #endregion


            #endregion

        }

        #endregion

        #region HELPERS

        /// <summary>
        /// Executes the make moves.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static object ExecuteMakeMoves(PmxDbConnection conn, object[] parameters)
        {
            PmxMoveProvider prov = new PmxMoveProvider(conn);

            string countedLocation = (string)parameters[0];
            Collection<int> listOfLuidsOnOtherLocation = (Collection<int>)parameters[1];

            Collection<int> listOfLuidsNotOnLocation = (Collection<int>)parameters[2];


            prov.MakeAdHocMoveForLUIDs(listOfLuidsOnOtherLocation, countedLocation, null, null);


            prov.MakeAdHocMoveForLUIDs(listOfLuidsNotOnLocation, Definitions.TEMP_LOCATION_CODE, null, null);



            return null;
        }



        /// <summary>
        /// Gets the number of rows.
        /// </summary>
        /// <param name="ds">The ds.</param>
        /// <returns></returns>
        private int GetNumberOfRows(DataSet ds)
        {
            if (ds != null && ds.Tables.Count > 0)
            {
                return ds.Tables[0].Rows.Count;
            }
            return 0;
        }


        /// <summary>
        /// Gets the new SSCC data set.
        /// </summary>
        /// <returns></returns>
        private DataSet GetNewSSCCDataSet()
        {
            DataSet ds = new DataSet();
            ds.Tables.Add(SSCCDataSet.NAME);
            ds.Tables[0].Columns.Add(SSCCDataSet.Columns.LUID.Name, SSCCDataSet.Columns.LUID.Types[1]);
            ds.Tables[0].Columns.Add(SSCCDataSet.Columns.SSCC.Name, SSCCDataSet.Columns.SSCC.Types[0]);
            ds.Tables[0].Columns.Add(SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name, SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Types[0]);

            return ds;

        }

        #endregion


        #region QUERY

        /// <summary>
        /// Static class grouping our native queries
        /// </summary>
        private class BuildQuery
        {

            /// <summary>
            /// Gets the SSCC info.
            /// </summary>
            /// <param name="sscc">The SSCC.</param>
            /// <returns></returns>
            public static string GetSSCCInfo(string sscc)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.Append("SELECT DISTINCT ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name);
                strBuilder.Append(',');
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME, SSCCDataSet.Columns.LUID.Name);
                strBuilder.Append(',');
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.SSCC.NAME, SSCCDataSet.Columns.SSCC.Name);

                strBuilder.AppendFormat("FROM {0} ", SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));

                strBuilder.AppendFormat("WHERE {0}.{1} = {2} ", SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME), SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.SSCC.NAME), SqlCommandHelper.FormatAndEscapeColumnValue(sscc, CultureInfo.InvariantCulture));

                return strBuilder.ToString();
            }



            /// <summary>
            /// Gets the SSCC info not on location.
            /// </summary>
            /// <param name="locationCode">The location code.</param>
            /// <param name="listOfCountedLUIDs">The list of counted LUID's.</param>
            /// <returns></returns>
            public static string GetSSCCInfoNotOnLocation(string locationCode, Collection<int> listOfCountedLUIDs)
            {
                StringBuilder strBuilder = new StringBuilder();

                strBuilder.Append("SELECT DISTINCT ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.StorLocCode.NAME, SSCCDataSet.Columns.STORAGE_LOCATION_CODE.Name);
                strBuilder.Append(',');
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME, SSCCDataSet.Columns.LUID.Name);
                strBuilder.Append(',');
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.SSCC.NAME, SSCCDataSet.Columns.SSCC.Name);

                strBuilder.AppendFormat("FROM {0} ", SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));

                strBuilder.AppendFormat("WHERE {0}.{1} = {2} ", SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME), SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.StorLocCode.NAME), SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));

                strBuilder.AppendFormat("AND {0}.{1} IS NOT NULL ", SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME), SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME));

                if (listOfCountedLUIDs != null &&
                    listOfCountedLUIDs.Count > 0)
                {
                    strBuilder.AppendFormat("AND  {0}.{1} NOT IN ( ", SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME), SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.LogUnitIdentKey.NAME), SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));

                    string comma = "";
                    foreach (int luid in listOfCountedLUIDs)
                    {
                        strBuilder.Append(comma);
                        strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(luid, CultureInfo.InvariantCulture));
                        comma = ", ";
                    }


                    strBuilder.Append(") ");
                }

                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets all the active location.
            /// </summary>
            /// <param name="warehouseCode">The warehouse code.</param>
            /// <param name="locationCode">The location code.</param>
            /// <returns>
            /// The sql statement to get all the active location
            /// </returns>
            public static string GetAllLocation(string warehouseCode, string locationCode)
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME, LocationDataSet.Columns.CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Name.NAME, LocationDataSet.Columns.DESCRIPTION.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME, LocationDataSet.Columns.WAREHOUSECODE.Name);

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(" INNER JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxOseGeneralTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseGeneralTable.NAME, PmxOseGeneralTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, PmxOseStorageLocationGeneralTable.NAME, PmxOseStorageLocationGeneralTable.Columns.Code.NAME);

                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.IsActive.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));

                if (!string.IsNullOrEmpty(warehouseCode))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.PmxWarehouseCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(warehouseCode, CultureInfo.InvariantCulture));
                }

                if (!string.IsNullOrEmpty(locationCode))
                {
                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseGeneralTable.Columns.Code.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(locationCode, CultureInfo.InvariantCulture));
                }

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxOseStorageLocationGeneralTable.Columns.Name.NAME));

                return strBuilder.ToString();
            }


        }

        #endregion
    }
}

